function [Input, Output] = Smp_Rate_Convert(uin,Tin,Tout,Switch)

%%% Input:
% uin: the input pulse defined on "tin";
%% Tin: Tin.mode and Tin.t MUST be assigned!!!
    % the input time settings:
    % Tin.mode: which value to use
    % 1. Tin.t: time axis; 
    % 2. Tin.tw: Time Window;
%% Tout: Tout.mode and Tout.N MUST be assigned!!!
    % the output time settings:
    % Tout.mode: which value to use: 
    % Tout.N: discretization of the output pulse;
    % 1. Tout.t and 2. Tout.tw, similar definition
    % 3. Tout.tw2tau: output TW devided by input pulse width;
    % Tout.N is the target discretization number
    
%%% Output:
% Out.U: the output pulse defined on "T";
% Out.tw: the output time window;
% Out.t: the output time axis;

uin = uin(:);
Nin = length(uin);

if strcmp(Tin.mode,'t')
    dt_in = Tin.t(2)-Tin.t(1);
    Tin.t = Tin.t(:);
    Tin.tw = dt_in*Nin;
elseif strcmp(Tin.mode,'tw')
    dt_in = Tin.tw/Nin;
    Tin.t = dt_in*(-Nin/2:Nin/2-1);
    Tin.t = Tin.t(:);
else
    error('Tin.mode are not properly assigned!!!');
end

uin = centering(uin,Tin.t);
tau = PlsWidth(uin,Tin.t);

%% Step 1: Check whether the parameters are complete
if isempty(Tout.mode)
    error('Tout.mode must be specified!!!');
end
if isempty(Tout.N)
    error('Tout.N MUST be assigned!!!');
end

TUin.t = Tin.t;
TUin.tw = Tin.tw;
TUin.U = uin;
Input = TUin;

if strcmp(Switch,'off')
    Output = Input;
    return;
end

%% Step 2: if the output is specified by tw2tau and Tout.N

if strcmp(Tout.mode,'tw2tau') % Overrule!!! the output is specified by tw2tau and Tout.N

    tw_tmp = tau*Tout.tw2tau;
    dt_in = Tin.t(2)-Tin.t(1);
    N_tmp = round(tw_tmp/dt_in);
    if mod(N_tmp,2) == 1
        N_tmp = N_tmp+1;
    end
    tw_tmp = N_tmp*dt_in;
    dt_out = tw_tmp/Tout.N;
    N_diff = abs(N_tmp-Nin);
    
    if tw_tmp >= Tin.tw % the required tw is great than that of u_in;        
        u_tmp = [zeros(N_diff/2,1); uin; zeros(N_diff/2,1)];  % we form a new pulse by zero padding on both sides of the pulse        
    else  % The required tw is less than that of u_in;
        u_tmp = uin((N_diff/2+1):(Nin-N_diff/2));
    end
    
    t_tmp = dt_in*(-N_tmp/2:N_tmp/2-1);
    w_tmp = fftshift(2*pi/tw_tmp*(-N_tmp/2:N_tmp/2-1));
    
    win = fftshift(2*pi/Tin.tw*(-Nin/2:Nin/2-1));
    ufin = fft(uin);
    
    uf_tmp = fft(u_tmp);
    if N_tmp >= Tout.N   % We need less points than N_tmp
        uf = uf_tmp([1:Tout.N/2 N_tmp-Tout.N/2+1:end]);
        Out.U = ifft(uf)*length(uf)/length(uf_tmp);
    elseif N_tmp < Tout.N  % More points are needed.
        Out.U = interpft(u_tmp,Tout.N);
    end    
    Out.tw = dt_out*Tout.N;
    Out.t = dt_out*(-Tout.N/2:Tout.N/2-1).';    
    Out.U = centering(Out.U,Out.t);
    
    TUin.t = Tin.t;
    TUin.tw = Tin.tw;
    TUin.U = uin;
    Input = TUin;
    Output = Out;
    return;
end

%% Step 3: Up sampling or down sampling

if strcmp(Tout.mode,'t')
    dt_out = Tout.t(2)-Tout.t(1);
    Tout.t = Tout.t(:);
    Tout.tw = dt_out*Tout.N;
elseif strcmp(Tout.mode,'tw')
    dt_out = Tout.tw/Tout.N;
    Tout.t = dt_out*(-Tout.N/2:Tout.N/2-1);
    Tout.t = Tout.t(:);
else
    error('Tout.mode are not properly assigned!!!');
end

TW2tw = Tout.tw/Tin.tw;

if TW2tw == 1
    if (Tout.N == Nin) % Do nothing;
        Out.U = uin;
        Out.tw = Tin.tw;
        Out.t = Tin.t;
    elseif (Tout.N > Nin) % Up interpolation
        [Out.U Out.t] = Up_interp(uin,Tin.t,Tout.N);
        Out.tw = (Out.t(2)-Out.t(1))*Tout.N;
    else % Down sampling
        [Out.U Out.t] = DownSampling(uin,Tin.t,Tout.N);
        Out.tw = (Out.t(2)-Out.t(1))*Tout.N;        
    end
    Out.U = centering(Out.U,Out.t);
%% Enlarge or Shrink the time window
else    
    dt_in = Tin.t(2)-Tin.t(1);
    tw_tmp = Tin.tw*TW2tw;          % the targeted TW width;
    N_tmp = round(tw_tmp/dt_in);    % round the number to an integer.   
    if mod(N_tmp,2) == 1
        N_tmp = N_tmp+1;            % how many points for the current discretization?
    end
    % form a new pulse based on the input discretization;
    N_diff = abs(N_tmp-Nin);      % how many points do we need to add or get rid of;
    if TW2tw > 1 % enlarge time window    
        u_tmp = [zeros(N_diff/2,1); uin; zeros(N_diff/2,1)]; % form a new pulse: more sample points; enlarged time window
    else
        u_tmp = uin((N_diff/2+1):(Nin-N_diff/2));
    end
    t_tmp = dt_in*(-N_tmp/2:N_tmp/2-1);
    
    % re-sample the pulse;
    if (N_tmp >= Tout.N)  % Down sampling
        [Out.U Out.t] = DownSampling(u_tmp,t_tmp,Tout.N);        
    else        % Up integration
        [Out.U Out.t] = Up_interp(u_tmp,t_tmp,Tout.N);        
    end
    Out.U = centering(Out.U,Out.t);
    Out.tw = (Out.t(2)-Out.t(1))*Tout.N;           
end        

Output = Out;
end

function [uout tout] = Up_interp(uin,tin,Ntarget)
N = length(uin);
smpseq = (-Ntarget/2:Ntarget/2-1).';
dt = (tin(2) - tin(1))/(Ntarget/N);
tout = dt*smpseq;
uout = centering(interpft(uin,Ntarget),tout);
end

function [uout tout] = DownSampling(uin,tin,Ntarget)
N = length(uin);
uf = ifft(uin);
Nsmp = Ntarget/2;
smpseq = [1:Nsmp  N-Nsmp+1:N];

tw = (tin(2)-tin(1))*N;
dt_out = tw/Ntarget;
naxis = (-Ntarget/2:Ntarget/2-1).';
tout = dt_out*naxis;
uout = centering(fft(uf(smpseq)),tout);
end

function out = centering(u,t) % to shift the center of pulse
N = length(u);
T = (t(2)-t(1))*N;
w = fftshift(2*pi/T*(-N/2:N/2-1).');
ua = abs(u).^2;
E = norm(u)^2;
tc = sum(t.*ua)/E;
out = ifft(exp(1i*w*tc).*fft(u));
end

function tau = PlsWidth (u,t)
% To compute the pulse width of "u"
ua = abs(u).^2;
E = norm(u)^2;
tc = sum(t.*ua)/E;
tau = 2*sqrt(sum((t-tc).^2.*ua)/E);
end