classdef hme_newton < hme_parameters
    
    properties
        options      % for solving the stationary solution
        NTout    % the output of newton solver
    end
    
    % Internal matrices
    properties (GetAccess='public',SetAccess='public')
        Df2
        Df2_fun
        dispD2
        gainD2
        I
    end
    
    methods
        function obj=hme_newton()
        end
        
        function setup(self,IS,del,sgm)
            
            self.N = length(IS.Uout);
            self.naxis = (-self.N/2:1:self.N/2-1).';
            self.T = IS.T;
            self.dt = self.T/self.N;
            self.t = IS.t;
            self.dw = 2*pi/self.T;                     % Freq resolution
            self.w = fftshift(self.naxis*self.dw);                         % Freq axis
            
            if nargin >= 4
                self.del = del;
                self.sgm = sgm;                
            end
            self.I = speye(self.N);
            self.options = optimset('Display','off','Jacobian','on','TolFun',1e-16,'TolX',...
                1e-14,'Algorithm','levenberg-marquardt','ScaleProblem','Jacobian',...
                'MaxIter',300,'LargeScale','on');
        end
        
        function [u0, phi0] = newton_solver(self,u_in,phi_in,TS)
%             figure(121),plot(self.t,abs(u_in)); hold on;
            [u0, phi0] = self.normalization(u_in,phi_in,TS);
%             figure(121),plot(self.t,abs(u0),'r--'); hold off;
%             fprintf('self.T = %.7g\n',self.T);
            gi0 = self.g0/(1 + norm(u0)^2/self.Psat);
            x0 = [real(u0);imag(u0);gi0;phi0];
            f = @(x0) self.HMEsetup(x0);

%             for ip = 2:5
%                 pw = 10^(-ip);
%                 figure(67);
%                 x0 = [real(u0);imag(u0);gi0;phi0];
%                 Dx = x0*pw;
%                 [F,J] = f(x0);
%                 Df = abs(f(x0+Dx) - F - J*Dx);
%                 plot(log10(Df)); hold on; grid on;
%             end
            
            [xout,F,exitflag,output] = fsolve (f,x0,self.options);
            
            vout = xout(1:self.N);
            wout = xout(self.N+1:2*self.N);
            Uout = vout + 1j*wout;
            phi = xout(end);
                       
            gi = xout(end-1);
            out.Uout = self.centering(Uout,self.t,self.w); % pulse shape
            out.phi = phi;   
            out.res = norm(F); % Residual of the function [ ideally, F = \vec(0) ]
            out.exitflag = exitflag; % the status of output.
            out.t = self.t;
            out.T = self.T;
            out.g_l = (gi - self.l);
            out.output = output;
            self.NTout = out;
            
        end
        
        function [F, J] = HMEsetup(self,x)
            vi = x(1:self.N);
            wi = x(self.N+1:self.N*2);
            gi = x(end-1);
            phi = x(end);
            
            Ei = norm(vi+1i*wi)^2;
            
            if 1
                vd2 = (self.Df2_fun(vi));
                wd2 = (self.Df2_fun(wi));
            else
                vd2 = self.Df2*vi;
                wd2 = self.Df2*wi;
            end
            
            vi2 = vi.^2;
            wi2 = wi.^2;
            ui2 = vi2 + wi2;
            ui4 = ui2.^2;
            viwi = vi.*wi;
            
            F1 = (gi-self.l).*vi + 2*phi*wi + self.b2*wd2 + gi/self.Omega^2*vd2 - 2*self.gam*ui2.*wi...
                + 2*(self.del*ui2 - self.sgm*ui4).*vi;
            F1v = gi*self.gainD2 + spdiags(-self.l - 4*self.gam*viwi + self.del*(6*vi2+2*wi2)...
                - 2*self.sgm*(ui4 + 4*vi2.*ui2),0,self.N,self.N);
            F1w = self.dispD2 + spdiags(2*phi - 2*self.gam*(vi2+3*wi2) + 4*self.del*viwi...
                - 8*self.sgm*viwi.*ui2,0,self.N,self.N);
            F1g = (vi+vd2/self.Omega^2);
            F1p = 2*wi;

%             F1 = 2*phi*wi + self.b2*wd2 - 2*self.gam*ui2.*wi;                           
%             F1 = self.b2*wd2;                                                   
%             F1v = zeros(self.N,self.N);
%             F1w = self.dispD2;
%             F1g = zeros(self.N,1);
%             F1p = zeros(self.N,1);
%             
%             F = F1;
%             J = [F1v,F1w,F1g,F1p];
            
            F2 = (gi-self.l).*wi - 2*phi*vi - self.b2*vd2 + gi/self.Omega^2*wd2 + 2*self.gam*ui2.*vi...
                + 2*(self.del*ui2 - self.sgm*ui4).*wi;            
            F2v = -self.dispD2 + spdiags(-2*phi + 2*self.gam*(3*vi2+wi2) + 4*self.del*viwi...
                - 8*self.sgm*viwi.*ui2,0,self.N,self.N);
            F2w = gi*self.gainD2 + spdiags(-self.l + 4*self.gam*viwi + 2*self.del*(vi2+3*wi2)...
                - 2*self.sgm*(ui4 + 4*wi2.*ui2),0,self.N,self.N);
            F2g = (wi+wd2/self.Omega^2);
            F2p = -2*vi;
            
%             F2 = 2*(-self.sgm*ui4).*wi;            
%             F2v = spdiags(- 8*self.sgm*viwi.*ui2,0,self.N,self.N);
%             F2w = spdiags(2*self.sgm*(ui4 + 4*wi2.*ui2),0,self.N,self.N);
%             F2g = zeros(self.N,1);
%             F2p = zeros(self.N,1);
%             F = F2;
%             J = [F2v,F2w,F2g,F2p];
            
            F3 = gi*(self.Psat + Ei) - self.g0*self.Psat;
            F3v = 2*gi*vi.';
            F3w = 2*gi*wi.';
            F3g = self.Psat + Ei;
            F3p = 0;
            
            F = [F1;F2;F3];
            J = [F1v, F1w, F1g, F1p; F2v, F2w, F2g, F2p; F3v, F3w, F3g, F3p];
            
        end
    end % Methods
    
    %% utilities
    methods
        function [U0,phi,TU_in,TU_out] = normalization(self,u0,phi,TS)                       
            Tin.mode = 'tw'; Tin.tw = self.T;
%             Tout.mode = TS.mode; Tout.N = TS.N; Tout.tw2tau = TS.tw2tau;
            Tout = TS;
            [TU_in, TU_out] = Smp_Rate_Convert(u0,Tin,Tout,'on');
            IS.Uout = TU_out.U;
            IS.T = TU_out.tw;
            IS.t = TU_out.t;
            self.setup(IS);
            U0 = self.phase_rot(TU_out.U);
            
            self.Df2_fun = @(x) (ifft(-self.w.^2.*fft(x)));
            if strcmp(TS.switch_FourierFD,'on')
                [~, self.Df2] = fourdif(self.N,2);
                self.Df2 = self.dw^2*self.Df2;
            else
                self.Df2 = sparseFD(7,2,self.dt,self.N);
            end
                        
            self.Psat = self.Esat/self.dt;       % Saturated power      
            self.dispD2 = self.b2*self.Df2;       % dispersion operator
            self.gainD2 = self.I + (1/self.Omega^2)*self.Df2;  %  Gain multiplier
        end
        
        function tau = PlsWidth (self,u)
            % To compute the pulse width of "u"
            ua = abs(u).^2;
            E = norm(u)^2;
            tc = sum(self.t.*ua)/E;
            tau = 2*sqrt(sum((self.t-tc).^2.*ua)/E);
        end
        
        function out = centering(self,u,t,w) % to shift the center of pulse
            u = full(u);
            if (nargin<3)
                t = self.t;
                w = self.w;
            end
            ua = abs(u).^2;
            E = norm(u)^2;
            tc = sum(t.*ua)/E;
            out = ifft(exp(1i*w*tc).*fft(u));
        end
    end
    
    methods (Static)
        function u_out = phase_rot(uin)
            [~, idx] = max(abs(uin));
            agl = angle(uin(idx));
            u_out = uin*exp(-1i*agl);
        end
    end
    
end % Classdef
